#!/usr/bin/python3
#  coding: utf-8
#
#       QtVtk.py
#
#       vtkをQt上で実現させるためのモジュール
#
#   20/05/24    新規作成
#      05/25    getPatchNamesBlockIndex:バグ修正
#               multiRegionで親のpolyMeshが存在しない場合、indexがずれる
#      08/02    getPatchNamesBlockIndex:multiRegionCaseのチェックを追加
#               修正前は、lagrangianがmultiRegionとして設定されてしまう。
#      09/15    vtkのimport状況、vtkのバージョンをを出力。
#      10/29    logFileCreaterのimportを追加。（logを出力させるため）
#   21/06/12    reloadRemakeRenderWindow:再読込用を追加
#      12/21    selectPatches:faceの両面を表示する様に修正。
#   22/01/01    vtk-9.1.0対応で大幅修正。
#      01/02    makeRenderWindow, reloadRemakeRenderWindow:引数として
#               selPatchNamesを追加。（vtk起動時に選択patchを表示させる為）
#


try:
    import logFileCreater
    logFileCreater.log()
except:
    pass

print("importing vtk...")
import vtk
print("  imported " + vtk.vtkVersion.GetVTKSourceVersion())
import QtVtkRenderWindowInteractor as Qvtk
from vtk.util.colors import *
import time
import glob
import os

#from PyQt4 import QtCore, QtGui
import getPyQtModule
PyQtModule = getPyQtModule.PyQtModule

#  import されているPyQtに応じてimportする
if PyQtModule == "PyQt5":
    from PyQt5.QtCore import *
    from PyQt5.QtGui import *
    from PyQt5.QtWidgets import *
elif PyQtModule == "PySide2":
    from PySide2.QtCore import *
    from PySide2.QtGui import *
    from PySide2.QtWidgets import *
elif PyQtModule == "PyQt4":
    from PyQt4.QtCore import *
    from PyQt4.QtGui import *
elif PyQtModule == "PySide":
    from PySide.QtCore import *
    from PySide.QtGui import *
else:
    print("import error: could not import PyQt4, PyQt5, PySide or PySide2!")
    exit()


#--------------------
#  vtkQWidget class
#--------------------
class vtkQWidget(Qvtk.QVTKRenderWindowInteractor):
    """ mouseの操作を「vtkInteractorStyleTrackballCamera()」に設定
    patchのselectSignalを定義
    したものにoverrideする。"""

    #signalを設定
    try:
        #PyQt系
        selectedSignal = pyqtSignal()       #patch選択時
        rightClickSignal = pyqtSignal()     #右クリック時
    except:
        #PySide系
        selectedSignal = Signal()           #patch選択時
        rightClickSignal = Signal()         #右クリック時

    def __init__(self, parent=None):
        super(vtkQWidget, self).__init__(parent)
        #super(vtkQWidget, self).__init__()
        inter = self.GetRenderWindow().GetInteractor()
        inter.SetInteractorStyle(vtk.vtkInteractorStyleTrackballCamera())



#--------------------------
#  OpenFoamPatchMesh class
#--------------------------
class OpenFoamPatchMesh():
    """ OpenFOAMのpatchをvtkを使って3D表示させ、各patchの境界条件を設定する。"""

    #--------------------------
    #  mouseSelectActor class
    #--------------------------
    class mouseSelectActor(vtk.vtkInteractorStyleTrackballCamera):
        """ mouseのeventStyleをoverrideする。"""

        def __init__(self, parent):
            self.AddObserver("LeftButtonPressEvent",self.leftButtonPressEvent)
            self.AddObserver("RightButtonPressEvent",self.rightButtonPressEvent)
            

            self.LastPickedActor = None
            self.LastPickedProperty = vtk.vtkProperty()

            self.actorNo = 0                #選択したactorNo
            self.oldTime = time.time()
            self.parent = parent

        #
        #  leftButtonPressEvent
        #-----------------------
        def leftButtonPressEvent(self, obj, event):
            """ 左クリックのeventをチェック。"""

            clickPos = self.GetInteractor().GetEventPosition()

            picker = vtk.vtkPropPicker()
            picker.Pick(clickPos[0], clickPos[1], 0, self.GetDefaultRenderer())

            # get the new
            self.NewPickedActor = picker.GetActor()

            #doubleClick?
            currTime = time.time()
            dTime = currTime - self.oldTime
            self.oldTime = currTime
            if dTime <= 0.5:
                # If something was selected
                if self.NewPickedActor:
                    # If we picked something before, reset its property
                    #if self.LastPickedActor:
                    #    self.LastPickedActor.GetProperty().DeepCopy(self.LastPickedProperty)
                    
                    # Save the property of the picked actor so that we can
                    # restore it next time
                    #self.LastPickedProperty.DeepCopy(self.NewPickedActor.GetProperty())

                    # Highlight the picked actor by changing its properties
                    # self.NewPickedActor.GetProperty().SetColor(1.0, 0.0, 1.0)
                    # self.NewPickedActor.GetProperty().SetDiffuse(1.0)
                    # self.NewPickedActor.GetProperty().SetSpecular(0.0)
                    
                    # save the last picked actor
                    #self.LastPickedActor = self.NewPickedActor

                    # actorのNoを取得
                    actors = self.GetDefaultRenderer().GetActors()
                    i = 0
                    for actor in actors:
                        if actor == self.NewPickedActor:
                            break
                        i += 1
                    self.actorNo = i

                    #ctrlKeyを確認
                    ctrl, shift = self.parent.vtkWidget._GetCtrlShift(event)
                    if ctrl == True:
                        #actor選択を追加する
                        self.parent.selectAddActor(self.actorNo)
                    else:
                        #actorを選択する
                        self.parent.selectActor(self.actorNo)

            self.OnLeftButtonDown()
            return

        #
        #  rightButtonPressEvent
        #------------------------
        def rightButtonPressEvent(self, obj, event):
            #self.parent.vtkWidget.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
            #self.parent.vtkWidget.setContextMenuPolicy(Qt.CustomContextMenu)
            #self.parent.vtkWidget.customContextMenuRequested.connect(self.parent.showPopupMenu)
            self.parent.vtkWidget.rightClickSignal.emit()
            #self.OnRightButtonDown()
            return


    def __init__(self, vtkWidget, caseDir, region=".", timeFolder="0"):
        self.caseDir = caseDir
        self.fileName = caseDir + "/system/controlDict"
        self.region = region
        self.timeFolder = timeFolder
        self.vtkWidget = vtkWidget  #vtkRenderWindowInteractor
        self.enabled_block = []     #表示する全patch名
        self.selected_block = []    #選択したpatch名
        self.current_block = []     #表示しているpatch名
        self.regionBlock = None

        self.showEdge = True        #edge表示/非表示flag
        self.setOpacity = False     #透明flag

        self.vtkWidget.setToolTip(_("ダブルクリックしてpatchを選択する"))

        #reader
        self.reader = vtk.vtkOpenFOAMReader()
        self.reader.SetFileName(self.fileName)
        #self.reader.Update()
        self.reader.UpdateTimeStep(float(self.timeFolder))
        self.reader.EnableAllPatchArrays()
        self.reader.Update()
        #patch名とblock_indexを取得
        self.getPatchNamesBlockIndex()

    #
    #  getPatchNamesBlockIndex
    def getPatchNamesBlockIndex(self):
        """
        regionのblockを取得
        """
        #blockDataを取得
        blockDataDict = {}
        output = self.reader.GetOutput()
        blockData = None
        #region名とdataを取得
        metaData = output.GetMetaData(0)
        blockName = self.getBlockName(metaData)
        if blockName == "internalMesh":
            blockData = output
            blockDataDict["."] = blockData
        else:
            blockData = output.GetBlock(0)
            blockDataDict["."] = blockData
            nBlocks = output.GetNumberOfBlocks()
            for i in range(1, nBlocks):
                metaData = output.GetMetaData(i)
                blockName = self.getBlockName(metaData)
                blockData = output.GetBlock(i)
                blockDataDict[blockName] = blockData
        #regionBlockを取得
        self.regionBlock = blockDataDict[self.region]

    #
    #  getBlockName
    def getBlockName(self, metaData):
        """ metaDataからそのblockのblockNameを取得する"""
        lines = str(metaData).split("\n")
        blockName = ""
        for line in lines:
            words = line.split()
            if words[0] == "NAME:":
                blockName = words[1]
                break
        return blockName

    #
    #  makeRenderWindow
    #-------------------
    def makeRenderWindow(self, selPatchNames):
        """ renderWindowを作成する。
        enabled_block[0]のpatchを選択表示させる。"""
        #patchのactorを作成
        self.actorDict = self.getRegionActors()
        self.current_block = list(self.actorDict.keys())
        self.enabled_block = self.current_block[:]
        #選択面をクリア
        #selPatchNames = []
        self.selectPatches(selPatchNames)
        # renderer
        self.ren = vtk.vtkRenderer()
        self.ren.GradientBackgroundOn()
        self.ren.SetBackground(1, 1, 1)
        self.ren.SetBackground2(0.4, 0.55, 0.75)
        #actorを追加
        for name in self.current_block:
            self.ren.AddActor(self.actorDict[name])

        renWin = self.vtkWidget.GetRenderWindow()
        renWin.NewInstance()
        renWin.AddRenderer(self.ren)

        #styleを設定(event)
        style = self.mouseSelectActor(self)
        style.SetDefaultRenderer(self.ren)
        interactor = renWin.GetInteractor()
        interactor.SetInteractorStyle(style)

        #add axis
        self.axesActor = vtk.vtkAxesActor()
        self.axes = vtk.vtkOrientationMarkerWidget()
        self.axes.SetOrientationMarker(self.axesActor)
        self.axes.SetInteractor(interactor)
        self.axes.EnabledOn()
        self.axes.InteractiveOn()
        self.ren.ResetCamera()

    #
    #  getRegionActors
    def getRegionActors(self):
        """ region内のpatchActorを取得"""
        regionBlock = self.regionBlock
        nBlocks = regionBlock.GetNumberOfBlocks()
        patchActorDict = {}
        for i in range(nBlocks):
            metaData = regionBlock.GetMetaData(i)
            blockName = self.getBlockName(metaData)
            #vtk 9.1.0から名称変更
            #           9.0.3以前               9.1.0以降
            if blockName == "Patches" or blockName == "boundary":
                patchBlock = regionBlock.GetBlock(i)
                patchNamesIds = self.getPatchNamesIds(patchBlock)
                patchActorDict = self.createPatchActors(regionBlock, patchNamesIds)
                break
        return patchActorDict

    #
    #  getPatchNamesIds
    def getPatchNamesIds(self, patchBlock):
        """ blockDataからpatchNameとpatchIdを取得する"""
        patchIds = []
        patchNames = []
        index = 3
        nBlocks = patchBlock.GetNumberOfBlocks()
        for i in range(nBlocks):
            metaData = patchBlock.GetMetaData(i)
            patchName = self.getBlockName(metaData)
            if self.region != ".":
                patchName = self.region + "/" + patchName
            patchIds.append(index)
            patchNames.append(patchName)
            print(patchName, index)
            index += 1
        return patchNames, patchIds

    #
    #  createPatchActors
    def createPatchActors(self, regionBlock, patchNamesIds):
        """ patchActor, outlineActor(featureEdge)を作成する"""
        (patchNames, patchIds) = patchNamesIds
        patchActorDict = {}
        for i in range(len(patchIds)):
            index = patchIds[i]
            extract_block = vtk.vtkExtractBlock()
            extract_block.SetInputData(regionBlock)
            extract_block.AddIndex(index)
            extract_block.Update()
            #filter
            geom_filter = vtk.vtkGeometryFilter()
            geom_filter.SetInputConnection(extract_block.GetOutputPort())
            #featureEdge
            feature_edge = vtk.vtkFeatureEdges()
            feature_edge.SetInputConnection(geom_filter.GetOutputPort())
            #mapper
            mapper = vtk.vtkCompositePolyDataMapper()
            mapper.SetInputConnection(geom_filter.GetOutputPort())
            mapper.ScalarVisibilityOff()
            edgeMapper = vtk.vtkCompositePolyDataMapper()
            edgeMapper.SetInputConnection(feature_edge.GetOutputPort())
            edgeMapper.ScalarVisibilityOff()
            #actor
            actor = vtk.vtkActor()
            actor.SetMapper(mapper)
            prop = actor.GetProperty()
            prop.SetColor(green)
            name = patchNames[i]
            patchActorDict[name] = actor
        return patchActorDict

    #
    #  deleteAllActors
    #------------------
    def deleteAllActors(self):
        """ 表示されている全てのactorを削除する"""
        #renderWindowから全actorを削除する
        for delActorName in self.actorDict.keys():
            if delActorName in self.current_block:
                self.ren.RemoveActor(self.actorDict[delActorName])
                self.current_block.remove(delActorName)

    #
    #  reloadRemakeRenderWindow
    #---------------------------
    def reloadRemakeRenderWindow(self, selPatchNames):
        """ reloadしてrenderWindowを再作成する。"""
        #actorを再作成
        self.actorDict = self.getRegionActors()
        self.current_block = list(self.actorDict.keys())
        self.enabled_block = self.current_block[:]
        #select表示させる（actorのpropertyを設定）
        #selPatchNames = []
        self.selectPatches(selPatchNames)
        #renderWindowにactorを追加
        for name in self.enabled_block:
            self.ren.AddActor(self.actorDict[name])
        #表示させる
        self.vtkWidget.Initialize()

    #
    #  selectActor
    #---------------
    def selectActor(self, actorNo):
        """ mouseのdoubleClickで選択したactorを選択表示に設定する。
        
        Args:
            actorNo (int)   :actorのindexNo"""
        selPatch = self.current_block[actorNo]
        selPatchNames = [selPatch]
        self.selectPatches(selPatchNames)

    #
    #  selectAddActor
    #-----------------
    def selectAddActor(self, actorNo):
        """ mouseのdoubleClickで選択したactorを選択表示に追加設定する。
        
        Args:
            actorNo (int)   :actorのindexNo"""
        selPatch = self.current_block[actorNo]
        selPatchNames = self.selected_block[:]
        selPatchNames.append(selPatch)
        self.selectPatches(selPatchNames)

    #
    #  selectPatches
    #----------------
    def selectPatches(self, selPatchNames):
        """
        指定したpatchを選択表示、それ以外を非選択表示に設定する。
        
        Args:
            selPatchNames (list(str))   :選択表示させるpatch名
        """
        for name in self.enabled_block:
            if name in selPatchNames:
                #選択部
                prop = self.actorDict[name].GetProperty()
                #prop.BackfaceCullingOn()
                #prop.FrontfaceCullingOn()
                #両面表示
                prop.BackfaceCullingOff()
                prop.FrontfaceCullingOff()
                prop.SetAmbient(0.5)
                prop.SetColor(red)
                if self.showEdge == True:
                    prop.EdgeVisibilityOn()
                else:
                    prop.EdgeVisibilityOff()
                if self.setOpacity == True:
                    prop.SetOpacity(0.5)
                else:
                    prop.SetOpacity(1.0)
                prop.SetEdgeColor(white)
                prop.SetRepresentationToSurface()
            else:
                #非選択部
                prop = self.actorDict[name].GetProperty()
                #prop.BackfaceCullingOn()
                #prop.FrontfaceCullingOn()
                #両面表示
                prop.BackfaceCullingOff()
                prop.FrontfaceCullingOff()
                prop.SetAmbient(0.3)
                prop.SetColor(green)
                if self.showEdge == True:
                    prop.EdgeVisibilityOn()
                else:
                    prop.EdgeVisibilityOff()
                #prop.EdgeVisibilityOff()
                if self.setOpacity == True:
                    prop.SetOpacity(0.5)
                else:
                    prop.SetOpacity(1.0)
                prop.SetEdgeColor(black)
                prop.SetRepresentationToSurface()
        self.selected_block = selPatchNames
        self.vtkWidget.selectedSignal.emit()

    #
    #  remakeRenderWindow
    #---------------------
    def remakeRenderWindow(self):
        """
        隠したpatchを表示し、元の形状を表示する。
        """
        #全actorを削除
        for actorName in self.current_block:
            self.ren.RemoveActor(self.actorDict[actorName])
        #全actorを表示
        for actorName in self.enabled_block:
            self.ren.AddActor(self.actorDict[actorName])
        self.current_block = self.enabled_block[:]
        # addPatches = list(set(self.enabled_block).difference(self.current_block))
        # for patch in addPatches:
        #     self.ren.AddActor(self.actorDict[patch])
        #     self.current_block.append(patch)

    #
    #  hidePatches
    #---------------
    def hidePatches(self, hidePatchNames):
        """ 指定したpatchを隠し、表示させない。
        """
        if len(self.current_block) <= 1:
            return
        for hideName in hidePatchNames:
            if hideName in self.current_block:
                self.ren.RemoveActor(self.actorDict[hideName])
                self.current_block.remove(hideName)

    #
    #  showPopupMenu
    #-----------------
    def showPopupMenu(self, pos):
        """ popupMenuを表示する。"""
        #menu = QtGui.QMenu()
        menu = QMenu()
        menu.addAction(_("edge表示/非表示の切替え"), self.onSwitchShowEdge)
        menu.addAction(_("透明/不透明の切替え"), self.onSwitchOpacity)
        menu.addAction(_("選択patchを隠す"), self.onHideSelectedPatches)
        menu.addAction(_("全patchを表示"), self.onShowAllPatches)
        menu.exec_(self.vtkWidget.mapToGlobal(pos))

    def onSwitchShowEdge(self):
        if self.showEdge == True:
            self.showEdge = False
        else:
            self.showEdge = True
        selectPatchNames = self.selected_block
        self.selectPatches(selectPatchNames)

    def onSwitchOpacity(self):
        if self.setOpacity == True:
            self.setOpacity = False
        else:
            self.setOpacity = True
        selectPatchNames = self.selected_block
        self.selectPatches(selectPatchNames)

    def onHideSelectedPatches(self):
        selectedPatches = self.selected_block
        self.hidePatches(selectedPatches)

    def onShowAllPatches(self):
        self.remakeRenderWindow()


if __name__ == "__main__":
    import gettext
    gettext.install("app")
    #_ = gettext.gettext
